#include <stdio.h>              /* C includes */
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "libcci/cci.h"         /* CCI includes */

#define VERSION "1.0"
#define STR_LEN 255

MCCIPort   ss_cci_port;
char       ss_script_file[STR_LEN], ss_host_name[STR_LEN];
char       *cmd_name;
long int   ss_port_num=0;
int        ss_no_sleep=0, ss_last_sleep=0, quiet_flag=0;

void ss_display_usage()
{
  printf("\n\nCCI Slide Show   %s\n\n", VERSION);
  printf("Usage: %s options\n\n  where options is one or more of:\n\n", cmd_name);
  printf("-file   followed by the name of a text file that contains one\n");
  printf("        instruction per line for the CCI-compliant browser.\n\n");
  printf("-host   followed by the host name of the machine running the browser.\n\n");
  printf("-port   follwed by the port number on which the browser is listening\n");
  printf("        for CCI messages.\n\n");
  printf("-url    follwed by a single URL to be fetched and displayed.\n\n");
  printf("-quiet  optional flag to disable run-time console messages.\n\n");
  exit(0);
}

int ss_make_cci_connection()  /* return 0 on success, 1 or 2 on failure */
{
  if (strcmp(ss_host_name, "default")==0)
    {
      printf("\n%s: Error, use -host to specify a host machine.\n", cmd_name);
      ss_display_usage();
    }
  if (ss_port_num==0)
    {
      printf("\n%s: Error, use -port to specify a port number.\n", cmd_name);
      ss_display_usage();
    }
  if (quiet_flag==0)
    printf("CCI Slide Show: Connecting to %s at port %d\n", ss_host_name,
	   ss_port_num);
  ss_cci_port=(MCCIPort) MCCIConnect(ss_host_name, ss_port_num, NULL, NULL);
  if (!ss_cci_port)
    {
      printf("\n%s: Error, could not connect to machine %s at port %d.\n",
	     cmd_name, ss_host_name, ss_port_num);
      ss_display_usage();
    }
}

ss_sleep(int sleep_time)
{
  if (quiet_flag==0)
    printf("CCI Slide Show: Sleeping for %d seconds.\n", sleep_time);
  if (sleep_time<0)
    sleep_time=0;
  sleep(sleep_time);
  ss_last_sleep=sleep_time;
}

int ss_show_next_url(char *url) /* return 0 on success, or 1 on failure */
{
  if (ss_no_sleep==1)
    ss_sleep(ss_last_sleep);
  if (!MCCIIsConnected(ss_cci_port))
    ss_make_cci_connection();
  
  if (quiet_flag==0)
    printf("CCI Slide Show: Getting URL %s\n", url);
  if (MCCIIsConnected(ss_cci_port))
    {
      MCCIGet(ss_cci_port, url, MCCI_DEFAULT, MCCI_ABSOLUTE, 0);
      return 0;
    }
  else
    return 1;
}

void ss_process_cmd(char s[STR_LEN])
{
  char *cmd, *var;
  int  free_flag=0;

  if ( (s[0]=='#') || (s[0]=='!') || (s[0]=='\n') ) /* # denotes a comment */
    return;
  cmd=strtok(s, " ");
  var=strtok(NULL, "");

  if (cmd==NULL)
    return;
  if (strlen(cmd)==0)
    return;
  if (var==NULL)           /* special case, only a URL on this line */
    {
      var=strdup(cmd);
      cmd=strdup("url");
      free_flag=1;
    }
  else 
    free_flag=0;
    
  if (var[strlen(var)-1]=='\n')
    var[strlen(var)-1]=NULL;
  
  if (strcmp(cmd, "url")==0)                                /* url */
    {
      if (ss_show_next_url(var)==1)
	printf("CCI Slide Show: Error, no connection.\n");
      ss_no_sleep=1;
    }

  if (strcmp(cmd, "sleep")==0)                              /* sleep */
    {
      ss_sleep(atoi(var));
      ss_no_sleep=0;
    }

  if (strcmp(cmd, "host")==0)                               /* host */
    {
      if (MCCIIsConnected(ss_cci_port))
	MCCIDisconnect(ss_cci_port);
      strcpy(ss_host_name, var);
    }

  if (strcmp(cmd, "port")==0)                               /* port */
    {
      if (MCCIIsConnected(ss_cci_port))
	MCCIDisconnect(ss_cci_port);
      ss_port_num = atoi(var);
    }

  if (strcmp(cmd, "quiet")==0)
    if (strcmp(var, "on")==0)
      quiet_flag=1;
    else
      quiet_flag=0;
  
  if (strcmp(cmd, "exec")==0)
    system(var);

  if (free_flag==1)
    {
      free(cmd);
      free(var);
    }
  return;
}

int ss_slide_show()   /* return 0 if file opens, 1 on failure */
{
  FILE *fp;
  char s[STR_LEN];

  if ( (fp=fopen(ss_script_file, "r")) == NULL)
    return 1;
  else
    {
      while (fgets(s, STR_LEN, fp)!=NULL)
	ss_process_cmd(s);
      fclose(fp);
      MCCIDisconnect(ss_cci_port);
      return 0;
    }
}

int ss_handle_args(int argc, char *argv[]) 
{
  int x=0, flag=0;

  if (argc==1)
    ss_display_usage();
  while (x<argc-1)
    {
      x++;

      if ( (strcmp(argv[x], "-file")==0) ||          /* -file */
	  (strcmp(argv[x], "-f")==0) )
	if (argc == x+1)
	  ss_display_usage();
	else
	  strcpy(ss_script_file, argv[x+1]);
      
      if ( (strcmp(argv[x], "-help")==0) ||          /* -help */
	  (strcmp(argv[x], "-h")==0) )
	ss_display_usage();
      
      if ( (strcmp(argv[x], "-version")==0) ||       /* -version */
	  (strcmp(argv[x], "-v")==0) )
	ss_display_usage();
      
      if (strcmp(argv[x], "-host")==0)               /* -host */
	if (argc == x+1)
	  ss_display_usage();
	else
	  strcpy(ss_host_name, argv[x+1]);

      if (strcmp(argv[x], "-port")==0)               /* -port */
	if (argc == x+1)
	  ss_display_usage();
	else
	  ss_port_num = atoi(argv[x+1]);

      if (strcmp(argv[x], "-url")==0)
	if (argc == x+1)
	  ss_display_usage();
	else
	  {
	    ss_process_cmd(argv[x+1]);
	    flag=1;
	  }
      
      if (strcmp(argv[x], "-quiet")==0)              /* -quiet */
	quiet_flag=1;
    }
  return flag;
}

main(int argc, char *argv[]) 
{
  cmd_name = argv[0];
  strcpy(ss_host_name, "default");
  strcpy(ss_script_file, "default");
  if (ss_handle_args(argc, argv)==0)
    {
      if (strcmp(ss_script_file, "default")==0)
	{
	  printf("\n%s: Error, use -file to specify a script file.\n", 
		 cmd_name);
	  ss_display_usage();
	}
      if (ss_slide_show()==1)
	{
	  printf("\n%s: Error, unable to read script file %s.\n", cmd_name, 
		 ss_script_file);
	  ss_display_usage();
	}
    }
}
